@@ -1,9 +1,14 @@
 import click
+import importlib.metadata
 from isoqmap.commands.download import download
 from isoqmap.commands.isoquan import isoquan
 from isoqmap.commands.isoqtl import isoqtl
 
+__version__ = importlib.metadata.version("isoqmap")
+
+
 @click.group()
+@click.version_option(__version__, prog_name="IsoQMap")
 def cli():
     """Isoform Quantification and QTL mapping"""
     pass
@@ -14,4 +19,3 @@ cli.add_command(isoqtl)
 
 if __name__ == '__main__':
     cli()
-
